#include <LiquidCrystal_I2C.h>
#include <SdFat.h>
#include <SPI.h>
#include <TMRpcm.h>

const int microphonePin = A0;
const int chipSelectPin = 3;

const int buttonPins[] = { 9, 8, 7};
const int numButtons = sizeof(buttonPins) / sizeof(buttonPins[0]);
volatile bool buttonPressed[numButtons] = { false };
const LiquidCrystal_I2C lcd(0x27, 16, 2);
SdFile lrc_file;
TMRpcm music;
char files[3][10];
int nr_files;
int current_file;
unsigned long song_time;
bool is_paused = true;
bool mic_on = true;
SdFat sd;
float score = 0.0f;

struct LyricsLine {
  unsigned long timestamp;
  char line[35];
};
LyricsLine currentLyric;
const char wavExt[] = ".wav";
const char lrcExt[] = ".lrc";

void setup() {
  lcd.init();
  lcd.backlight();

  Serial.begin(9600);

  if (!sd.begin(chipSelectPin, SPI_FULL_SPEED)) {
    Serial.println("SD card initialization failed!");
    while (1);
  }

  music.speakerPin = 10;
  music.setVolume(5);
  music.quality(1);

  listFiles();

  current_file = nr_files - 1;

  for (int i = 0; i < numButtons; i++) {
    pinMode(buttonPins[i], INPUT_PULLUP);
    attachInterrupt(digitalPinToInterrupt(buttonPins[i]), buttonInterruptHandler, FALLING);
  }

  cli();

  // Setup Timer2 for 10ms
  TCCR2A = 0;
  TCCR2B = 0;
  TCNT2 = 0;
  OCR2A = 155;
  TCCR2A |= (1 << WGM21);
  TCCR2B |= (1 << CS22) | (1 << CS21) | (1 << CS20);
  TIMSK2 |= (1 << OCIE2A);

  sei();
  
  lcd.clear();
  displayTextOnLCD("Initialization done. Press Next!");
}

ISR(TIMER2_COMPA_vect) {
  if (!is_paused) {
    if (mic_on) {
      int readMicrophone = analogRead(microphonePin);
      score += (float) readMicrophone / 1024;
    }
    song_time += 10;
  }
}

void loop() {

  for (int i = 0; i < numButtons; i++) {
    int buttonState = digitalRead(buttonPins[i]);

    if (buttonState == LOW && !buttonPressed[i]) {
      delay(50);
      if (digitalRead(buttonPins[i]) == LOW) {
        Serial.print("Button ");
        Serial.print(i);
        Serial.println(" pressed");
        performAction(i);

        buttonPressed[i] = true;
      }
    } else if (buttonState == HIGH && buttonPressed[i]) {
      buttonPressed[i] = false;
    }
  }

  if (!is_paused) {
    if (song_time >= currentLyric.timestamp) {
      displayTextOnLCD(currentLyric.line);
      music.pause();
      readLyricsLine();
      music.pause();
    } 
  }

  if (Serial.available() > 0) {
    char command = Serial.read();
    if (command = 'S') {
      Serial.print("Score: ");
      Serial.println(score);
    }
  }
}

void buttonInterruptHandler() {
  static bool attached = false;
  if (!attached) {
    for (int i = 0; i < numButtons; i++) {
      if (digitalRead(buttonPins[i]) == LOW) {
        buttonPressed[i] = true;
        delay(50);
        while (digitalRead(buttonPins[i]) == LOW) {
        }
      }
    }
  }
}

void displayTextOnLCD(const char* text) {
  lcd.clear();

  char textBuffer[17];

  if (strlen(text) <= 16) {
    lcd.setCursor(0, 0);
    lcd.print(text);
  } else {
    lcd.setCursor(0, 0);
    strncpy(textBuffer, text, 16);
    textBuffer[16] = '\0';
    lcd.print(textBuffer);

    lcd.setCursor(0, 1);
    strncpy(textBuffer, text + 16, 16);
    textBuffer[16] = '\0';
    lcd.print(textBuffer);
  }
}

bool isWAV(const char* filename) {
  const char* fileExt = strrchr(filename, '.');
  if (fileExt != nullptr) {
    return strcmp(fileExt + 1, "wav") == 0;
  }

  return false;
}

void listFiles() {
  SdFile root;
  SdFile entry;

  if (!root.open("/", O_RDONLY)) {
    Serial.println("Failed to open root directory!");
    return;
  }

  while (1) {
      if (!entry.openNext(&root, O_RDONLY)) {
        break;
      }

      char filename[10];

      entry.getName(filename, sizeof(filename) - 1);
      filename[sizeof(filename) - 1] = '\0';

      entry.close();

      Serial.print("File: ");
      Serial.println(filename);

      if (!isWAV(filename)) {
        continue;
      }

      strncpy(files[nr_files], filename, strlen(filename) - 4);
      files[nr_files][strlen(filename) - 4] = '\0';
      nr_files++;
  }

  root.close();
}

void readLyricsLine() {
  char timestampStr[9], sizeTimestamp = 0, sizeLine = 0;
  memset(currentLyric.line, '\0', sizeof(currentLyric.line));
  currentLyric.timestamp = 0;

  if (lrc_file.available()) {
    char c = lrc_file.read();
    while (c != '\n' && c != '\r' && c != -1) {
      if (c == '[') {
        c = lrc_file.read();
        while (c != ']') {
          timestampStr[sizeTimestamp++] = c;
          c = lrc_file.read();
        }
      } else {
        currentLyric.line[sizeLine++] = c;
      }
      
      c = lrc_file.read();
    }

    currentLyric.timestamp = parseTimestamp(timestampStr);
  } 
}

unsigned long parseTimestamp(const char* timestampStr) {
  unsigned long minutes = (timestampStr[0] - '0') * 10 + (timestampStr[1] - '0');
  unsigned long seconds = (timestampStr[3] - '0') * 10 + (timestampStr[4] - '0');
  unsigned long milliseconds = (timestampStr[6] - '0') * 10 + (timestampStr[7] - '0');
  
  return ((minutes * 60 + seconds) * 1000) + milliseconds;
}

void performAction(int buttonIndex) {
  if (buttonIndex == 2) {
    current_file = (current_file + 1) % nr_files;
    char filename[10];

    music.disable();

    strcpy(filename, files[current_file]);
    strcat(filename, lrcExt);
    lrc_file.close();

    if (!lrc_file.open(filename, O_READ)) {
      Serial.println(filename);
      Serial.println("Failed to open lrc");
      while (1);
    }

    readLyricsLine();

    strcpy(filename, files[current_file]);
    strcat(filename, wavExt);
    music.play(filename);

    song_time = 0;

    is_paused = false;
  } else if (buttonIndex == 1) {
    is_paused = !is_paused;
    music.pause();
  } else if (buttonIndex == 0) {
    mic_on = !mic_on;
    if (mic_on) {
      Serial.println("Mic on");
    } else {
      Serial.println("Mic off");
    }
  }
}
